@import url("styling.css");

.greeter flowboxchild {
  -gtk-outline-radius: 9px;
}

.greeter flowboxchild:hover {
  -gtk-icon-effect: highlight;
}

flowboxchild:hover {
  background-color: @theme_selected_bg_color;
}

flowboxchild:hover * {
  color: @theme_selected_fg_color;
}

.info:hover {
  color: @theme_selected_fg_color;
  background-color: @theme_selected_bg_color;
}
